@@ -15,7 +15,12 @@
  * See the License for the specific language governing permissions a* limitations under the License.
  */
 
-import { Clipboard, Dialog, showDialog } from '@jupyterlab/apputils';
+import {
+  Clipboard,
+  Dialog,
+  InputDialog,
+  showDialog,
+} from '@jupyterlab/apputils';
 import { PathExt } from '@jupyterlab/coreutils';
 import { DocumentWidget } from '@jupyterlab/docregistry';
 import { FileEditor } from '@jupyterlab/fileeditor';
@@ -25,18 +30,13 @@ import {
   addIcon,
   pythonIcon,
   fileIcon,
-  rKernelIcon
+  rKernelIcon,
+  markdownIcon,
 } from '@jupyterlab/ui-components';
 import { CodeEditor, IEditorServices } from '@jupyterlab/codeeditor';
 import * as nbformat from '@jupyterlab/nbformat';
 import { JupyterFrontEnd } from '@jupyterlab/application';
-import {
-  Cell,
-  CodeCellModel,
-  ICodeCellModel,
-  MarkdownCell,
-  CodeCell
-} from '@jupyterlab/cells';
+import { CodeCellModel, MarkdownCell, CodeCell } from '@jupyterlab/cells';
 
 import { Widget } from '@lumino/widgets';
 import { find, StringExt } from '@lumino/algorithm';
@@ -44,15 +44,12 @@ import { Drag } from '@lumino/dragdrop';
 import { MimeData } from '@lumino/coreutils';
 
 import React from 'react';
+import { CodeSnippetService, ICodeSnippet } from './CodeSnippetService';
+import { FilterTools } from './CodeSnippetFilterTools';
+import { showPreview } from './CodeSnippetPreview';
+import { showMoreOptions } from './CodeSnippetMenu';
 
-import { CodeSnippetWidgetModel } from './CodeSnippetWidgetModel';
-import { FilterTools } from './FilterTools';
-import { showPreview } from './PreviewSnippet';
-import { showMoreOptions } from './MoreOptions';
-import {
-  ICodeSnippet,
-  CodeSnippetContentsService
-} from './CodeSnippetContentsService';
+import { CodeSnippetContentsService } from './CodeSnippetContentsService';
 
 import moreSVGstr from '../style/icon/jupyter_moreicon.svg';
 import {
@@ -92,8 +89,10 @@ import {
   sbtIcon,
   rustIcon,
   qsharpIcon,
-  sasIcon
+  sasIcon,
 } from './CodeSnippetLanguages';
+import { ICodeSnippetEditorMetadata } from './CodeSnippetEditor';
+import { showMessage } from './CodeSnippetMessage';
 
 /**
  * The CSS class added to code snippet widget.
@@ -119,7 +118,11 @@ const CODE_SNIPPET_MORE_OTPIONS_COPY = 'jp-codeSnippet-more-options-copy';
 const CODE_SNIPPET_MORE_OTPIONS_INSERT = 'jp-codeSnippet-more-options-insert';
 const CODE_SNIPPET_MORE_OTPIONS_EDIT = 'jp-codeSnippet-more-options-edit';
 const CODE_SNIPPET_MORE_OTPIONS_DELETE = 'jp-codeSnippet-more-options-delete';
+const CODE_SNIPPET_MORE_OTPIONS_DOWNLOAD =
+  'jp-codeSnippet-more-options-download';
 const CODE_SNIPPET_CREATE_NEW_BTN = 'jp-createSnippetBtn';
+const CODE_SNIPPET_NAME = 'jp-codeSnippet-name';
+const OPTIONS_BODY = 'jp-codeSnippet-options-body';
 
 /**
  * The threshold in pixels to start a drag event.
@@ -141,7 +144,7 @@ const JUPYTER_CELL_MIME = 'application/vnd.jupyter.cells';
  */
 const moreOptionsIcon = new LabIcon({
   name: 'custom-ui-components:moreOptions',
-  svgstr: moreSVGstr
+  svgstr: moreSVGstr,
 });
 
 /**
@@ -149,22 +152,22 @@ const moreOptionsIcon = new LabIcon({
  */
 interface ICodeSnippetDisplayProps {
   codeSnippets: ICodeSnippet[];
+  codeSnippetManager: CodeSnippetService;
   app: JupyterFrontEnd;
   getCurrentWidget: () => Widget;
-  openCodeSnippetEditor: (args: any) => void;
+  openCodeSnippetEditor: (args: ICodeSnippetEditorMetadata) => void;
   editorServices: IEditorServices;
-  _codeSnippetWidgetModel: CodeSnippetWidgetModel;
-  updateCodeSnippets: () => void;
+  updateCodeSnippetWidget: () => void;
 }
 
 /**
  * CodeSnippetDisplay state.
  */
 interface ICodeSnippetDisplayState {
-  codeSnippets: ICodeSnippet[];
-  matchIndices: number[][];
   searchValue: string;
   filterTags: string[];
+  selectedLangTags: string[];
+  searchOptions: string[];
 }
 
 /**
@@ -179,19 +182,131 @@ export class CodeSnippetDisplay extends React.Component<
   constructor(props: ICodeSnippetDisplayProps) {
     super(props);
     this.state = {
-      codeSnippets: this.props.codeSnippets,
-      matchIndices: [],
       searchValue: '',
-      filterTags: []
+      filterTags: [],
+      selectedLangTags: [],
+      searchOptions: [],
     };
     this._drag = null;
     this._dragData = null;
     this.handleDragMove = this.handleDragMove.bind(this);
     this._evtMouseUp = this._evtMouseUp.bind(this);
     this.handleRenameSnippet = this.handleRenameSnippet.bind(this);
+    this.setSearchOptions = this.setSearchOptions.bind(this);
   }
 
-  // Handle code snippet insert into an editor
+  filterSnippets = (
+    codeSnippets: ICodeSnippet[],
+    searchValue: string,
+    filterTags: string[],
+    selectedLangTags: string[]
+  ): {
+    filteredCodeSnippets: ICodeSnippet[];
+    matchedIndices: { [id: number]: number[] };
+  } => {
+    // filter with search
+    let filteredSnippets = codeSnippets.slice();
+    const matchedIndices: { [id: number]: number[] } = {};
+
+    if (searchValue !== '') {
+      const matchResults: StringExt.IMatchResult[] = [];
+      const filteredSnippetsScore: {
+        score: number;
+        snippet: ICodeSnippet;
+      }[] = [];
+
+      // language, title, code
+      filteredSnippets.forEach((snippet) => {
+        const matchResult = StringExt.matchSumOfSquares(
+          (
+            snippet.language +
+            snippet.name +
+            snippet.code.join('\n')
+          ).toLowerCase(),
+          searchValue.replace(' ', '').toLowerCase()
+        );
+        if (matchResult) {
+          matchResults.push(matchResult);
+          filteredSnippetsScore.push({
+            score: matchResult.score,
+            snippet: snippet,
+          });
+        }
+      });
+
+      // sort snippets by its score
+      filteredSnippetsScore.sort((a, b) => a.score - b.score);
+      const newFilteredSnippets: ICodeSnippet[] = [];
+      filteredSnippetsScore.forEach((snippetScore) =>
+        newFilteredSnippets.push(snippetScore.snippet)
+      );
+      filteredSnippets = newFilteredSnippets;
+
+      // sort the matchResults by its score
+      matchResults.sort((a, b) => a.score - b.score);
+
+      matchResults.forEach((res, id) => {
+        matchedIndices[filteredSnippets[id].id] = res.indices;
+      });
+    }
+
+    // filter with tags
+    if (filterTags.length !== 0) {
+      filteredSnippets = filteredSnippets.filter((codeSnippet) => {
+        return filterTags.some((tag) => {
+          if (codeSnippet.tags) {
+            if (selectedLangTags.length !== 0) {
+              // lang tags selected
+              if (
+                codeSnippet.tags.includes(tag) &&
+                selectedLangTags.includes(codeSnippet.language)
+              ) {
+                return true;
+              }
+              //if only language tags are selected
+              else if (
+                filterTags.length === selectedLangTags.length &&
+                filterTags.every((value) => selectedLangTags.includes(value))
+              ) {
+                if (selectedLangTags.includes(codeSnippet.language)) {
+                  return true;
+                }
+              }
+            } else {
+              // no lang tags selected
+              if (codeSnippet.tags.includes(tag)) {
+                return true;
+              }
+            }
+          }
+          return false;
+        });
+      });
+    }
+    // find id's that are not in filteredSnippets
+    const willBeRemovedIds = [];
+    for (const key in matchedIndices) {
+      let hasKey = false;
+      for (const codeSnippet of filteredSnippets) {
+        if (codeSnippet.id === parseInt(key)) {
+          hasKey = true;
+        }
+      }
+      if (hasKey === false) {
+        willBeRemovedIds.push(parseInt(key));
+      }
+    }
+
+    // if the snippet does not have the tag, remove its mathed index
+    willBeRemovedIds.forEach((id: number) => delete matchedIndices[id]);
+
+    return {
+      filteredCodeSnippets: filteredSnippets,
+      matchedIndices: matchedIndices,
+    };
+  };
+
+  // Handle code snippet insert into a notebook or document
   private insertCodeSnippet = async (snippet: ICodeSnippet): Promise<void> => {
     const widget: Widget = this.props.getCurrentWidget();
     const snippetStr: string = snippet.code.join('\n');
@@ -277,7 +392,7 @@ export class CodeSnippetDisplay extends React.Component<
         '" is incompatible with ' +
         editorLanguage +
         '. Continue?',
-      buttons: [Dialog.cancelButton(), Dialog.okButton()]
+      buttons: [Dialog.cancelButton(), Dialog.okButton()],
     });
   };
 
@@ -286,23 +401,25 @@ export class CodeSnippetDisplay extends React.Component<
     return showDialog({
       title: 'Error',
       body: errMsg,
-      buttons: [Dialog.okButton()]
+      buttons: [Dialog.okButton()],
     });
   };
 
   // Create 6 dots drag/drop image on hover
   private dragHoverStyle = (id: number): void => {
     document
-      .getElementsByClassName(CODE_SNIPPET_DRAG_HOVER)
-      [id].classList.add(CODE_SNIPPET_DRAG_HOVER_SELECTED);
+      .querySelector(`#${CODE_SNIPPET_DRAG_HOVER}${id}`)
+      .classList // .getElementsByClassName(CODE_SNIPPET_DRAG_HOVER)
+      // [id].classList.
+      .add(CODE_SNIPPET_DRAG_HOVER_SELECTED);
   };
 
   // Remove 6 dots off hover
   private dragHoverStyleRemove = (id: number): void => {
-    if (document.getElementsByClassName(CODE_SNIPPET_DRAG_HOVER)) {
+    if (document.getElementsByClassName(CODE_SNIPPET_DRAG_HOVER_SELECTED)) {
       document
-        .getElementsByClassName(CODE_SNIPPET_DRAG_HOVER)
-        [id].classList.remove(CODE_SNIPPET_DRAG_HOVER_SELECTED);
+        .querySelector(`#${CODE_SNIPPET_DRAG_HOVER}${id}`)
+        .classList.remove(CODE_SNIPPET_DRAG_HOVER_SELECTED);
     }
   };
 
@@ -310,80 +427,89 @@ export class CodeSnippetDisplay extends React.Component<
   private boldNameOnSearch = (
     id: number,
     language: string,
-    name: string
+    name: string,
+    matchedIndices: number[]
   ): JSX.Element => {
     const displayName = language + name;
 
     // check if the searchValue is not ''
     if (this.state.searchValue !== '') {
       const elements = [];
-      const boldIndices = this.state.matchIndices[id].slice();
-
-      // get first match index in the name
-      let i = 0;
-      while (i < boldIndices.length) {
-        if (boldIndices[i] >= language.length) {
-          elements.push(displayName.substring(language.length, boldIndices[i]));
-          break;
-        }
-        i++;
-      }
-
-      // when there is no match in name but language
-      if (i >= boldIndices.length) {
-        return <span>{name}</span>;
-      } else {
-        // current and next indices are bold indices
-        let currIndex = boldIndices[i];
-        let nextIndex;
-        // check if the match is the end of the name
-        if (i < boldIndices.length - 1) {
+      if (matchedIndices) {
+        // get first match index in the name
+        let i = 0;
+        while (i < matchedIndices.length) {
+          if (matchedIndices[i] >= language.length) {
+            elements.push(
+              displayName.substring(language.length, matchedIndices[i])
+            );
+            break;
+          }
           i++;
-          nextIndex = boldIndices[i];
-        } else {
-          nextIndex = null;
         }
-        while (nextIndex !== null) {
-          // make the current index bold
-          elements.push(
-            <mark key={id + '_' + currIndex} className={SEARCH_BOLD}>
-              {displayName.substring(currIndex, currIndex + 1)}
-            </mark>
-          );
-          // add the regular string until we reach the next bold index
-          elements.push(displayName.substring(currIndex + 1, nextIndex));
-          currIndex = nextIndex;
-          if (i < boldIndices.length - 1) {
+
+        // when there is no match in name but language
+        if (i >= matchedIndices.length) {
+          return <span>{name}</span>;
+        } else {
+          // current and next indices are bold indices
+          let currIndex = matchedIndices[i];
+          let nextIndex;
+          // check if the match is the end of the name
+          if (i < matchedIndices.length - 1) {
             i++;
-            nextIndex = boldIndices[i];
+            nextIndex = matchedIndices[i];
           } else {
             nextIndex = null;
           }
+          while (nextIndex !== null) {
+            // make the current index bold
+            elements.push(
+              <mark key={id + '_' + currIndex} className={SEARCH_BOLD}>
+                {displayName.substring(currIndex, currIndex + 1)}
+              </mark>
+            );
+            // add the regular string until we reach the next bold index
+            elements.push(displayName.substring(currIndex + 1, nextIndex));
+            currIndex = nextIndex;
+            if (i < matchedIndices.length - 1) {
+              i++;
+              nextIndex = matchedIndices[i];
+            } else {
+              nextIndex = null;
+            }
+          }
+          if (nextIndex === null) {
+            elements.push(
+              <mark key={id + '_' + currIndex} className={SEARCH_BOLD}>
+                {displayName.substring(currIndex, currIndex + 1)}
+              </mark>
+            );
+            elements.push(
+              displayName.substring(currIndex + 1, displayName.length)
+            );
+          }
+          return <span>{elements}</span>;
         }
-        if (nextIndex === null) {
-          elements.push(
-            <mark key={id + '_' + currIndex} className={SEARCH_BOLD}>
-              {displayName.substring(currIndex, currIndex + 1)}
-            </mark>
-          );
-          elements.push(
-            displayName.substring(currIndex + 1, displayName.length)
-          );
-        }
-        return <span>{elements}</span>;
       }
     }
-    return <span onDoubleClick={this.handleRenameSnippet}>{name}</span>;
+    return (
+      <span
+        title={'Double click to rename'}
+        className={CODE_SNIPPET_NAME}
+        onDoubleClick={this.handleRenameSnippet}
+      >
+        {name}
+      </span>
+    );
   };
 
   // rename snippet on double click
-  // TODO: duplicate name check!
   private async handleRenameSnippet(
     event: React.MouseEvent<HTMLSpanElement, MouseEvent>
   ): Promise<void> {
-    const contentsService = CodeSnippetContentsService.getInstance();
     const target = event.target as HTMLElement;
-    const oldPath = 'snippets/' + target.innerHTML + '.json';
+    const oldName = target.innerHTML;
 
     const new_element = document.createElement('input');
     new_element.setAttribute('type', 'text');
@@ -398,30 +524,40 @@ export class CodeSnippetDisplay extends React.Component<
 
     new_element.onblur = async (): Promise<void> => {
       if (target.innerHTML !== new_element.value) {
-        const newPath = 'snippets/' + new_element.value + '.json';
-        try {
-          await contentsService.rename(oldPath, newPath);
-        } catch (error) {
-          new_element.replaceWith(target);
+        const newName = new_element.value;
 
+        const isDuplicateName = this.props.codeSnippetManager.duplicateNameExists(
+          newName
+        );
+
+        if (isDuplicateName) {
           await showDialog({
             title: 'Duplicate Name of Code Snippet',
-            body: <p> {`"${newPath}" already exists.`} </p>,
-            buttons: [Dialog.okButton({ label: 'Dismiss' })]
+            body: <p> {`"${newName}" already exists.`} </p>,
+            buttons: [Dialog.okButton({ label: 'Dismiss' })],
           });
-          return;
+        } else {
+          this.props.codeSnippetManager
+            .renameSnippet(oldName, newName)
+            .then(async (res: boolean) => {
+              if (res) {
+                target.innerHTML = new_element.value;
+              } else {
+                console.log('Error in renaming snippet!');
+              }
+            });
         }
-        this.props._codeSnippetWidgetModel.renameSnippet(
-          target.innerHTML,
-          new_element.value
-        );
-        target.innerHTML = new_element.value;
       }
       new_element.replaceWith(target);
     };
     new_element.onkeydown = (event: KeyboardEvent): void => {
       switch (event.code) {
-        case 'Enter' || 'NumpadEnter': // Enter
+        case 'Enter': // Enter
+          event.stopPropagation();
+          event.preventDefault();
+          new_element.blur();
+          break;
+        case 'NumpadEnter': // Enter
           event.stopPropagation();
           event.preventDefault();
           new_element.blur();
@@ -463,7 +599,7 @@ export class CodeSnippetDisplay extends React.Component<
     this._dragData = {
       pressX: event.clientX,
       pressY: event.clientY,
-      dragImage: target.nextSibling.firstChild.cloneNode(true) as HTMLElement
+      dragImage: target.nextSibling.firstChild.cloneNode(true) as HTMLElement,
     };
 
     const dragImageTextColor = getComputedStyle(document.body).getPropertyValue(
@@ -510,8 +646,13 @@ export class CodeSnippetDisplay extends React.Component<
         event.clientY
       )
     ) {
-      const idx = (event.target as HTMLElement).id;
-      const codeSnippet = this.state.codeSnippets[parseInt(idx)];
+      const idx = (event.target as HTMLElement).id.slice(
+        CODE_SNIPPET_DRAG_HOVER.length
+      );
+
+      const codeSnippet = this.props.codeSnippets.filter(
+        (codeSnippet) => codeSnippet.id === parseInt(idx)
+      )[0];
 
       void this.startDrag(
         data.dragImage,
@@ -550,8 +691,7 @@ export class CodeSnippetDisplay extends React.Component<
   ): Promise<void> {
     const target = event.target as HTMLElement;
 
-    const modelFactory = new ModelFactory();
-    const model = modelFactory.createCodeCell({});
+    const model = new CodeCellModel({});
     model.value.text = codeSnippet.code.join('\n');
     model.metadata;
 
@@ -562,7 +702,7 @@ export class CodeSnippetDisplay extends React.Component<
       dragImage: dragImage,
       supportedActions: 'copy-move',
       proposedAction: 'copy',
-      source: this
+      source: this,
     });
 
     this._drag.mimeData.setData(JUPYTER_CELL_MIME, selected);
@@ -592,8 +732,9 @@ export class CodeSnippetDisplay extends React.Component<
   //Set the position of the preview to be next to the snippet title.
 
   private _setPreviewPosition(id: number): void {
-    const realTarget = document.getElementsByClassName(TITLE_CLASS)[id];
-    const newTarget = document.getElementsByClassName(CODE_SNIPPET_ITEM)[id];
+    const realTarget = document.querySelector(`#${TITLE_CLASS}${id}`);
+    const newTarget = document.querySelector(`#${CODE_SNIPPET_ITEM}${id}`);
+    // (CODE_SNIPPET_ITEM)[id];
     // distDown is the number of pixels to shift the preview down
     const distDown: number = realTarget.getBoundingClientRect().top - 43; //this is bumping it up
     const elementSnippet = newTarget as HTMLElement;
@@ -603,16 +744,13 @@ export class CodeSnippetDisplay extends React.Component<
       '--preview-max-height',
       heightPreview
     );
-    // if (realTarget.getBoundingClientRect().top > window.screen.height / 2) {
-    //   distDown = distDown - 66; //this is bumping it up further if it's close to the end of the screen
-    // }
     const final = distDown.toString(10) + 'px';
     document.documentElement.style.setProperty('--preview-distance', final);
   }
 
   //Set the position of the option to be under to the three dots on snippet.
   private _setOptionsPosition(
-    event: React.MouseEvent<HTMLButtonElement, MouseEvent>
+    event: React.MouseEvent<HTMLElement, MouseEvent>
   ): void {
     const target = event.target as HTMLElement;
     let top: number;
@@ -625,8 +763,10 @@ export class CodeSnippetDisplay extends React.Component<
       top -= 120;
     }
     const leftAsString =
-      target.getBoundingClientRect().left.toString(10) + 'px';
+      (target.parentElement.style.left + event.pageX).toString() + 'px';
+
     const topAsString = top.toString(10) + 'px';
+
     document.documentElement.style.setProperty(
       '--more-options-top',
       topAsString
@@ -651,6 +791,18 @@ export class CodeSnippetDisplay extends React.Component<
           />
         );
       }
+      case 'Gfm': {
+        return (
+          <markdownIcon.react
+            tag="span"
+            height="16px"
+            width="16px"
+            right="7px"
+            top="5px"
+            margin-right="3px"
+          />
+        );
+      }
       case 'Java': {
         return (
           <javaIcon.react
@@ -1125,8 +1277,9 @@ export class CodeSnippetDisplay extends React.Component<
   // Render display of code snippet list
   private renderCodeSnippet = (
     codeSnippet: ICodeSnippet,
-    id: number
+    matchedIndices: number[]
   ): JSX.Element => {
+    const id = codeSnippet.id;
     const buttonClasses = BUTTON_CLASS;
     const displayName = '[' + codeSnippet.language + '] ' + codeSnippet.name;
     const name = codeSnippet.name;
@@ -1136,30 +1289,40 @@ export class CodeSnippetDisplay extends React.Component<
       {
         title: 'Insert, copy, edit, and delete',
         icon: moreOptionsIcon,
-        onClick: (
-          event: React.MouseEvent<HTMLButtonElement, MouseEvent>
-        ): void => {
-          showMoreOptions({ body: new OptionsHandler(this, codeSnippet) });
+        onClick: (event: React.MouseEvent<HTMLElement, MouseEvent>): void => {
+          showMoreOptions({
+            body: new OptionsHandler(this, codeSnippet),
+          });
           this._setOptionsPosition(event);
-        }
-      }
+        },
+      },
     ];
     return (
       <div
         key={codeSnippet.name}
         className={CODE_SNIPPET_ITEM}
-        id={id.toString()}
+        id={`${CODE_SNIPPET_ITEM}${id}`}
+        title={'Right click for more options'}
         onMouseOver={(): void => {
           this.dragHoverStyle(id);
         }}
         onMouseOut={(): void => {
           this.dragHoverStyleRemove(id);
         }}
+        onContextMenu={(
+          event: React.MouseEvent<HTMLElement, MouseEvent>
+        ): void => {
+          event.preventDefault();
+          showMoreOptions({
+            body: new OptionsHandler(this, codeSnippet),
+          });
+          this._setOptionsPosition(event);
+        }}
       >
         <div
           className={CODE_SNIPPET_DRAG_HOVER}
           title="Drag to move"
-          id={id.toString()}
+          id={`${CODE_SNIPPET_DRAG_HOVER}${id}`}
           onMouseDown={(event): void => {
             this.handleDragSnippet(event);
           }}
@@ -1172,7 +1335,7 @@ export class CodeSnippetDisplay extends React.Component<
                 id: id,
                 title: displayName,
                 body: new PreviewHandler(),
-                codeSnippet: codeSnippet
+                codeSnippet: codeSnippet,
               },
               this.props.editorServices
             );
@@ -1182,13 +1345,17 @@ export class CodeSnippetDisplay extends React.Component<
             this._evtMouseLeave();
           }}
         >
-          <div key={displayName} className={TITLE_CLASS} id={id.toString()}>
-            <div id={id.toString()} title={name} className={DISPLAY_NAME_CLASS}>
+          <div
+            key={displayName}
+            className={TITLE_CLASS}
+            id={`${TITLE_CLASS}${id}`}
+          >
+            <div id={id.toString()} className={DISPLAY_NAME_CLASS}>
               {this.renderLanguageIcon(language)}
-              {this.boldNameOnSearch(id, language, name)}
+              {this.boldNameOnSearch(id, language, name, matchedIndices)}
             </div>
             <div className={ACTION_BUTTONS_WRAPPER_CLASS} id={id.toString()}>
-              {actionButtons.map(btn => {
+              {actionButtons.map((btn) => {
                 return (
                   <button
                     key={btn.title}
@@ -1211,122 +1378,39 @@ export class CodeSnippetDisplay extends React.Component<
               })}
             </div>
           </div>
-          <div className={CODE_SNIPPET_DESC} id={id.toString()}>
-            <p id={id.toString()}>{`${codeSnippet.description}`}</p>
-          </div>
+          {this.renderDescription(codeSnippet, id)}
         </div>
       </div>
     );
   };
 
-  static getDerivedStateFromProps(
-    props: ICodeSnippetDisplayProps,
-    state: ICodeSnippetDisplayState
-  ): ICodeSnippetDisplayState {
-    if (state.searchValue === '' && state.filterTags.length === 0) {
-      return {
-        codeSnippets: props.codeSnippets,
-        matchIndices: [],
-        searchValue: '',
-        filterTags: []
-      };
+  renderDescription(codeSnippet: ICodeSnippet, id: number): JSX.Element {
+    if (codeSnippet.description.length !== 0) {
+      return (
+        <div className={CODE_SNIPPET_DESC} id={id.toString()}>
+          <p id={id.toString()}>{`${codeSnippet.description}`}</p>
+        </div>
+      );
+    } else {
+      return null;
     }
-
-    if (state.searchValue !== '' || state.filterTags.length !== 0) {
-      // const newSnippets = props.codeSnippets.filter(codeSnippet => {
-      //   return (
-      //     state.matchIndices[codeSnippet.id] !== null ||
-      //     // (state.searchValue !== '' &&
-      //     //   codeSnippet.name.toLowerCase().includes(state.searchValue)) ||
-      //     // (state.searchValue !== '' &&
-      //     //   codeSnippet.language.toLowerCase().includes(state.searchValue)) ||
-      //     (codeSnippet.tags &&
-      //       codeSnippet.tags.some(tag => state.filterTags.includes(tag)))
-      //   );
-      // });
-      return {
-        codeSnippets: state.codeSnippets,
-        matchIndices: state.matchIndices,
-        searchValue: state.searchValue,
-        filterTags: state.filterTags
-      };
-    }
-    return null;
   }
 
-  filterSnippets = (searchValue: string, filterTags: string[]): void => {
-    // filter with search
-    let matchIndices: number[][] = [];
-    const matchResults: StringExt.IMatchResult[] = [];
-    let filteredSnippets = this.props.codeSnippets;
-    const filteredSnippetsScore: {
-      score: number;
-      snippet: ICodeSnippet;
-    }[] = [];
-    if (searchValue !== '') {
-      filteredSnippets.forEach(snippet => {
-        const matchResult = StringExt.matchSumOfSquares(
-          (snippet.language + snippet.name).toLowerCase(),
-          searchValue.replace(' ', '').toLowerCase()
-        );
-
-        if (matchResult) {
-          matchResults.push(matchResult);
-          filteredSnippetsScore.push({
-            score: matchResult.score,
-            snippet: snippet
-          });
-        }
-      });
-
-      // sort snippets by its score
-      filteredSnippetsScore.sort((a, b) => a.score - b.score);
-      const newFilteredSnippets: ICodeSnippet[] = [];
-      filteredSnippetsScore.forEach(snippetScore =>
-        newFilteredSnippets.push(snippetScore.snippet)
-      );
-      filteredSnippets = newFilteredSnippets;
-
-      // sort the matchResults by its score
-      matchResults.sort((a, b) => a.score - b.score);
-      matchResults.forEach(res => matchIndices.push(res.indices));
-    }
-
-    // filter with tags
-    if (filterTags.length !== 0) {
-      const newMatchIndices = matchIndices.slice();
-      filteredSnippets = filteredSnippets.filter((codeSnippet, id) => {
-        return filterTags.some(tag => {
-          if (codeSnippet.tags) {
-            if (codeSnippet.tags.includes(tag)) {
-              return true;
-            }
-          }
-          // if the snippet does not have the tag, remove its mathed index
-          const matchedIndex = matchIndices[id];
-          const indexToDelete = newMatchIndices.indexOf(matchedIndex);
-          newMatchIndices.splice(indexToDelete, 1);
-          return false;
-        });
-      });
-      matchIndices = newMatchIndices;
-    }
-
-    this.setState(
-      {
-        codeSnippets: filteredSnippets,
-        matchIndices: matchIndices,
-        searchValue: searchValue,
-        filterTags: filterTags
-      },
-      () => {
-        console.log('snippets filtered');
-      }
-    );
+  handleFilter = (
+    searchValue: string,
+    filterTags: string[],
+    selectedLangTags: string[]
+  ): void => {
+    this.setState({
+      searchValue: searchValue,
+      filterTags: filterTags,
+      selectedLangTags: selectedLangTags,
+    });
   };
 
-  getActiveTags(): string[] {
+  getActiveTags(): string[][] {
     const tags: string[] = [];
+    const languages: string[] = [];
     for (const codeSnippet of this.props.codeSnippets) {
       if (codeSnippet.tags) {
         for (const tag of codeSnippet.tags) {
@@ -1335,22 +1419,47 @@ export class CodeSnippetDisplay extends React.Component<
           }
         }
       }
+      if (!languages.includes(codeSnippet.language)) {
+        languages.push(codeSnippet.language);
+      }
     }
-    return tags;
+    return [tags, languages];
   }
 
+  getActiveTagsDictionary = (): Map<string, string[]> => {
+    const tagsAndLangs: Map<string, string[]> = new Map<string, string[]>();
+    for (const codeSnippet of this.props.codeSnippets) {
+      if (codeSnippet.tags) {
+        // check if tag is in dict, if it is add lang to value (if not already present)
+        // if tag not in dict add tag as key and lang as first val
+        for (const tag of codeSnippet.tags) {
+          if (tag !== codeSnippet.language) {
+            if (tagsAndLangs.has(tag)) {
+              const langs = tagsAndLangs.get(tag);
+              if (!langs.includes(codeSnippet.language)) {
+                langs.push(codeSnippet.language);
+              }
+              tagsAndLangs.set(tag, langs);
+            } else {
+              tagsAndLangs.set(tag, [codeSnippet.language]);
+            }
+          }
+        }
+      }
+    }
+    return tagsAndLangs;
+  };
+
   private deleteCommand(codeSnippet: ICodeSnippet): void {
-    const contentsService = CodeSnippetContentsService.getInstance();
     showDialog({
       title: 'Delete snippet?',
       body: 'Are you sure you want to delete "' + codeSnippet.name + '"? ',
       buttons: [
-        Dialog.okButton({
+        Dialog.cancelButton(),
+        Dialog.warnButton({
           label: 'Delete',
-          displayType: 'warn'
         }),
-        Dialog.cancelButton()
-      ]
+      ],
     }).then((response: any): void => {
       if (response.button.accept) {
         const widgetId = `${CODE_SNIPPET_EDITOR}-${codeSnippet.id}`;
@@ -1360,24 +1469,54 @@ export class CodeSnippetDisplay extends React.Component<
             return widget.id === widgetId;
           }
         );
+
         if (editor) {
           editor.dispose();
         }
-
         // deleting snippets when there is one snippet active
-        contentsService.delete('snippets/' + codeSnippet.name + '.json');
-        this.props._codeSnippetWidgetModel.deleteSnippet(codeSnippet.id);
-        this.props._codeSnippetWidgetModel.reorderSnippet();
-        this.props._codeSnippetWidgetModel.updateSnippetContents();
+        this.props.codeSnippetManager
+          .deleteSnippet(codeSnippet.id)
+          .then((result: boolean) => {
+            if (result) {
+              this.props.updateCodeSnippetWidget();
+            } else {
+              console.log('Error in deleting the snippet');
+              return;
+            }
+          });
+      }
+    });
+  }
 
-        // active tags after delete
-        const activeTags = this.getActiveTags();
+  private downloadCommand(codeSnippet: ICodeSnippet): void {
+    // Request a text
+    InputDialog.getText({
+      title: 'Download Snippet?',
+      label: 'Directory to Download: ',
+      placeholder: 'share/snippet',
+      okLabel: 'Download',
+    }).then((value: Dialog.IResult<string>) => {
+      if (value.button.accept) {
+        const dirs = value.value.split('/');
 
-        // filterTags: only the tags that are still being used
-        this.setState(state => ({
-          codeSnippets: this.props._codeSnippetWidgetModel.snippets,
-          filterTags: state.filterTags.filter(tag => activeTags.includes(tag))
-        }));
+        const codeSnippetDownloader = CodeSnippetContentsService.getInstance();
+
+        let path = '';
+        for (let i = 0; i < dirs.length; i++) {
+          path += dirs[i] + '/';
+          codeSnippetDownloader.save(path, { type: 'directory' }).catch((_) => {
+            alert('Path should be a relative path');
+          });
+        }
+
+        path += codeSnippet.name + '.json';
+
+        codeSnippetDownloader.save(path, {
+          type: 'file',
+          format: 'text',
+          content: JSON.stringify(codeSnippet),
+        });
+        showMessage('download');
       }
     });
   }
@@ -1393,6 +1532,7 @@ export class CodeSnippetDisplay extends React.Component<
   // create dropdown menu
   public createOptionsNode(codeSnippet: ICodeSnippet): HTMLElement {
     const body = document.createElement('div');
+    body.className = OPTIONS_BODY;
 
     const optionsContainer = document.createElement('div');
     optionsContainer.className = CODE_SNIPPET_MORE_OTPIONS_CONTENT;
@@ -1408,23 +1548,25 @@ export class CodeSnippetDisplay extends React.Component<
     copySnip.textContent = 'Copy snippet to clipboard';
     copySnip.onclick = (): void => {
       Clipboard.copyToSystem(codeSnippet.code.join('\n'));
-      alert('saved to clipboard');
+      showMessage('copy');
       this.removeOptionsNode();
     };
     const editSnip = document.createElement('div');
     editSnip.className = CODE_SNIPPET_MORE_OTPIONS_EDIT;
     editSnip.textContent = 'Edit snippet';
     editSnip.onclick = (): void => {
-      const allTags = this.getActiveTags();
+      const allSnippetTags = this.getActiveTags()[0]; // snippet tags only
+      const allLangTags = this.getActiveTags()[1];
       this.props.openCodeSnippetEditor({
         name: codeSnippet.name,
         description: codeSnippet.description,
         language: codeSnippet.language,
         code: codeSnippet.code,
         id: codeSnippet.id,
-        selectedTags: codeSnippet.tags,
-        allTags: allTags,
-        fromScratch: false
+        selectedTags: codeSnippet.tags, // snippet tags
+        allSnippetTags: allSnippetTags,
+        allLangTags: allLangTags,
+        fromScratch: false,
       });
       this.removeOptionsNode();
     };
@@ -1435,15 +1577,38 @@ export class CodeSnippetDisplay extends React.Component<
       this.deleteCommand(codeSnippet);
       this.removeOptionsNode();
     };
+
+    const downloadSnip = document.createElement('div');
+    downloadSnip.className = CODE_SNIPPET_MORE_OTPIONS_DOWNLOAD;
+    downloadSnip.textContent = 'Download snippet';
+    downloadSnip.onclick = (): void => {
+      this.downloadCommand(codeSnippet);
+      this.removeOptionsNode();
+    };
+
     optionsContainer.appendChild(insertSnip);
     optionsContainer.appendChild(copySnip);
     optionsContainer.appendChild(editSnip);
+    optionsContainer.appendChild(downloadSnip);
     optionsContainer.appendChild(deleteSnip);
     body.append(optionsContainer);
     return body;
   }
 
+  setSearchOptions(selectedOptions: string[]): void {
+    this.setState({
+      searchOptions: selectedOptions,
+    });
+  }
+
   render(): React.ReactElement {
+    const { filteredCodeSnippets, matchedIndices } = this.filterSnippets(
+      this.props.codeSnippets,
+      this.state.searchValue,
+      this.state.filterTags,
+      this.state.selectedLangTags
+    );
+
     return (
       <div>
         <header className={CODE_SNIPPETS_HEADER_CLASS}>
@@ -1456,9 +1621,11 @@ export class CodeSnippetDisplay extends React.Component<
                 description: '',
                 language: 'Python',
                 code: [],
-                id: -1,
-                allTags: this.getActiveTags(),
-                fromScratch: true
+                id: this.props.codeSnippets.length,
+                selectedTags: [],
+                allSnippetTags: this.getActiveTags()[0],
+                allLangTags: this.getActiveTags()[1],
+                fromScratch: true,
               });
             }}
           >
@@ -1466,13 +1633,18 @@ export class CodeSnippetDisplay extends React.Component<
           </button>
         </header>
         <FilterTools
-          tags={this.getActiveTags()}
-          onFilter={this.filterSnippets}
+          tagDictionary={this.getActiveTagsDictionary()}
+          languageTags={this.getActiveTags()[1]}
+          snippetTags={this.getActiveTags()[0]}
+          onFilter={this.handleFilter}
         />
         <div className={CODE_SNIPPETS_CONTAINER}>
           <div>
-            {this.state.codeSnippets.map((codeSnippet, id) =>
-              this.renderCodeSnippet(codeSnippet, id)
+            {filteredCodeSnippets.map((codeSnippet) =>
+              this.renderCodeSnippet(
+                codeSnippet,
+                matchedIndices[codeSnippet.id]
+              )
             )}
           </div>
         </div>
@@ -1505,40 +1677,3 @@ class Private {
     return this.createPreviewContent();
   }
 }
-
-/**
- * A content factory for console children.
- */
-export interface IContentFactory extends Cell.IContentFactory {
-  /**
-   * Create a new code cell widget.
-   */
-  createCodeCell(options: CodeCell.IOptions): CodeCell;
-}
-
-/**
- * The default implementation of an `IModelFactory`.
- */
-export class ModelFactory {
-  /**
-   * The factory for output area models.
-   */
-  readonly codeCellContentFactory: CodeCellModel.IContentFactory;
-
-  /**
-   * Create a new code cell.
-   *
-   * @param source - The data to use for the original source data.
-   *
-   * @returns A new code cell. If a source cell is provided, the
-   *   new cell will be initialized with the data from the source.
-   *   If the contentFactory is not provided, the instance
-   *   `codeCellContentFactory` will be used.
-   */
-  createCodeCell(options: CodeCellModel.IOptions): ICodeCellModel {
-    if (!options.contentFactory) {
-      options.contentFactory = this.codeCellContentFactory;
-    }
-    return new CodeCellModel(options);
-  }
-}
