@@ -20,19 +20,20 @@ import {
   ReactWidget,
   showDialog,
   Dialog,
-  WidgetTracker
+  WidgetTracker,
 } from '@jupyterlab/apputils';
 import { Button } from '@jupyterlab/ui-components';
-import { Contents } from '@jupyterlab/services';
 
 import { Message } from '@lumino/messaging';
 
 import React from 'react';
 
-import { CodeSnippetContentsService } from './CodeSnippetContentsService';
+import { CodeSnippetService } from './CodeSnippetService';
 import { CodeSnippetWidget } from './CodeSnippetWidget';
 import { SUPPORTED_LANGUAGES } from './CodeSnippetLanguages';
 import { CodeSnippetEditorTags } from './CodeSnippetEditorTags';
+import { showMessage } from './CodeSnippetMessage';
+import { validateInputs, saveOverWriteFile } from './CodeSnippetUtilities';
 
 /**
  * CSS style classes
@@ -49,11 +50,6 @@ const CODE_SNIPPET_EDITOR_MIRROR = 'jp-codeSnippetInput-editor';
 const CODE_SNIPPET_EDITOR_INPUTAREA = 'jp-codeSnippetInputArea';
 const CODE_SNIPPET_EDITOR_INPUTAREA_MIRROR = 'jp-codeSnippetInputArea-editor';
 
-const CODE_SNIPPET_EDITOR_INPUTNAME_VALIDITY =
-  'jp-codeSnippet-inputName-validity';
-const CODE_SNIPPET_EDITOR_INPUTDESC_VALIDITY =
-  'jp-codeSnippet-inputDesc-validity';
-
 const EDITOR_DIRTY_CLASS = 'jp-mod-dirty';
 
 export interface ICodeSnippetEditorMetadata {
@@ -63,7 +59,8 @@ export interface ICodeSnippetEditorMetadata {
   code: string[];
   id: number;
   selectedTags: string[];
-  allTags: string[];
+  allSnippetTags: string[];
+  allLangTags: string[];
   fromScratch: boolean;
 }
 
@@ -74,21 +71,19 @@ export class CodeSnippetEditor extends ReactWidget {
   private oldCodeSnippetName: string;
   private _codeSnippetEditorMetaData: ICodeSnippetEditorMetadata;
   private _hasRefreshedSinceAttach: boolean;
-  contentsService: CodeSnippetContentsService;
+  contentsService: CodeSnippetService;
   codeSnippetWidget: CodeSnippetWidget;
   tracker: WidgetTracker;
 
   constructor(
-    contentsService: CodeSnippetContentsService,
     editorServices: IEditorServices,
     tracker: WidgetTracker,
     codeSnippetWidget: CodeSnippetWidget,
     args: ICodeSnippetEditorMetadata
   ) {
     super();
-
     this.addClass(CODE_SNIPPET_EDITOR);
-    this.contentsService = contentsService;
+    this.contentsService = CodeSnippetService.getCodeSnippetService();
     this.editorServices = editorServices;
     this.tracker = tracker;
 
@@ -193,9 +188,9 @@ export class CodeSnippetEditor extends ReactWidget {
           value: this._codeSnippetEditorMetaData.code.join('\n'),
           mimeType: getMimeTypeByLanguage({
             name: this._codeSnippetEditorMetaData.language,
-            codemirror_mode: this._codeSnippetEditorMetaData.language
-          })
-        })
+            codemirror_mode: this._codeSnippetEditorMetaData.language,
+          }),
+        }),
       });
       this.editor.model.value.changed.connect((args: any) => {
         this._codeSnippetEditorMetaData.code = args.text.split('\n');
@@ -219,7 +214,7 @@ export class CodeSnippetEditor extends ReactWidget {
       this.update();
     }
 
-    window.addEventListener('beforeunload', e => {
+    window.addEventListener('beforeunload', (e) => {
       if (!this.saved) {
         e.preventDefault();
         e.returnValue = '';
@@ -254,10 +249,9 @@ export class CodeSnippetEditor extends ReactWidget {
         buttons: [
           Dialog.cancelButton(),
           Dialog.warnButton({ label: 'Discard' }),
-          Dialog.okButton({ label: 'Save' })
-        ]
+          Dialog.okButton({ label: 'Save' }),
+        ],
       }).then((response: any): void => {
-        console.log(response.button);
         if (response.button.accept) {
           if (response.button.label === 'Discard') {
             this.dispose();
@@ -273,9 +267,9 @@ export class CodeSnippetEditor extends ReactWidget {
               `.${CODE_SNIPPET_EDITOR}-${this._codeSnippetEditorMetaData.id} .${CODE_SNIPPET_EDITOR_LANG_INPUT}`
             ) as HTMLSelectElement).value;
 
-            const validity = this.validateInputs(name, description, language);
+            const validity = validateInputs(name, description, language);
             if (validity) {
-              this.updateSnippet().then(value => {
+              this.updateSnippet().then((value) => {
                 if (value) {
                   this.dispose();
                   super.onCloseRequest(msg);
@@ -349,49 +343,12 @@ export class CodeSnippetEditor extends ReactWidget {
       `.${CODE_SNIPPET_EDITOR}-${this._codeSnippetEditorMetaData.id} .${CODE_SNIPPET_EDITOR_LANG_INPUT}`
     ) as HTMLSelectElement).value;
 
-    const validity = this.validateInputs(name, description, language);
+    const validity = validateInputs(name, description, language);
     if (validity) {
       this.updateSnippet();
     }
   }
 
-  private validateInputs(
-    name: string,
-    description: string,
-    language: string
-  ): boolean {
-    let status = true;
-    let message = '';
-    if (name === '') {
-      message += 'Name must be filled out\n';
-      status = false;
-    }
-    if (name.match(/[^a-z0-9_]+/)) {
-      message += 'Wrong format of the name\n';
-      status = false;
-    }
-    if (description === '') {
-      message += 'Description must be filled out\n';
-      status = false;
-    }
-    if (description.match(/[^a-zA-Z0-9_ ,.?!]+/)) {
-      message += 'Wrong format of the description\n';
-      status = false;
-    }
-    if (language === '') {
-      message += 'Language must be filled out';
-      status = false;
-    }
-    if (!SUPPORTED_LANGUAGES.includes(language)) {
-      message += 'Language must be one of the options';
-      status = false;
-    }
-    if (status === false) {
-      alert(message);
-    }
-    return status;
-  }
-
   async updateSnippet(): Promise<boolean> {
     const name = (document.querySelector(
       `.${CODE_SNIPPET_EDITOR}-${this._codeSnippetEditorMetaData.id} .${CODE_SNIPPET_EDITOR_NAME_INPUT}`
@@ -407,64 +364,75 @@ export class CodeSnippetEditor extends ReactWidget {
     this._codeSnippetEditorMetaData.description = description;
     this._codeSnippetEditorMetaData.language = language;
 
-    const newPath =
-      'snippets/' + this._codeSnippetEditorMetaData.name + '.json';
+    const newName = this._codeSnippetEditorMetaData.name;
+    const oldName = this.oldCodeSnippetName;
 
-    if (!this._codeSnippetEditorMetaData.fromScratch) {
-      const oldPath = 'snippets/' + this.oldCodeSnippetName + '.json';
+    const newSnippet = {
+      name: this._codeSnippetEditorMetaData.name,
+      description: this._codeSnippetEditorMetaData.description,
+      language: this._codeSnippetEditorMetaData.language,
+      code: this._codeSnippetEditorMetaData.code,
+      id: this._codeSnippetEditorMetaData.id,
+      tags: this._codeSnippetEditorMetaData.selectedTags,
+    };
 
-      if (newPath !== oldPath) {
-        // renaming code snippet
-        try {
-          await this.contentsService.rename(oldPath, newPath);
-        } catch (error) {
-          await showDialog({
-            title: 'Duplicate Name of Code Snippet',
-            body: <p> {`"${newPath}" already exists.`} </p>,
-            buttons: [Dialog.okButton({ label: 'Dismiss' })]
-          });
-          return false;
-        }
+    const isDuplicatName = this.contentsService.duplicateNameExists(newName);
 
-        // set new name as an old name
-        this.oldCodeSnippetName = this._codeSnippetEditorMetaData.name;
-      }
-    } else {
-      let nameCheck = false;
-      await this.contentsService
-        .getData(newPath, 'file')
-        .then(async (value: Contents.IModel) => {
-          if (value.name) {
-            await showDialog({
-              title: 'Duplicate Name of Code Snippet',
-              body: <p> {`"${newPath}" already exists.`} </p>,
-              buttons: [Dialog.okButton({ label: 'Dismiss' })]
-            });
+    // set new name as an old name
+    this.oldCodeSnippetName = this._codeSnippetEditorMetaData.name;
+
+    // add new snippet
+    if (this._codeSnippetEditorMetaData.fromScratch) {
+      if (isDuplicatName) {
+        const oldSnippet = this.contentsService.getSnippet(newName)[0];
+        await saveOverWriteFile(this.contentsService, oldSnippet, newSnippet);
+      } else {
+        this.contentsService.addSnippet(newSnippet).then((res: boolean) => {
+          if (!res) {
+            console.log('Error in adding snippet');
+            return false;
           }
-        })
-        .catch(() => {
-          nameCheck = true;
         });
-      if (!nameCheck) {
-        return false;
+        showMessage('confirm');
       }
     }
+    // modify existing snippet
+    else {
+      if (newName !== oldName) {
+        if (isDuplicatName) {
+          // overwrite
+          const oldSnippet = this.contentsService.getSnippet(newName)[0];
+          await saveOverWriteFile(
+            this.contentsService,
+            oldSnippet,
+            newSnippet
+          ).then((res: boolean) => {
+            if (res) {
+              // get the id of snippet you are editting
+              const removedSnippet = this.contentsService.getSnippet(
+                oldName
+              )[0];
+
+              // delete the one you are editting
+              this.contentsService.deleteSnippet(removedSnippet.id);
+            } else {
+              return false;
+            }
+          });
+        }
+      }
+      this.contentsService
+        .modifyExistingSnippet(oldName, newSnippet)
+        .then((res: boolean) => {
+          if (!res) {
+            console.log('Error in modifying snippet');
+            return false;
+          }
+        });
+    }
 
     this.saved = true;
 
-    await this.contentsService.save(newPath, {
-      type: 'file',
-      format: 'text',
-      content: JSON.stringify({
-        name: this._codeSnippetEditorMetaData.name,
-        description: this._codeSnippetEditorMetaData.description,
-        language: this._codeSnippetEditorMetaData.language,
-        code: this._codeSnippetEditorMetaData.code,
-        id: this._codeSnippetEditorMetaData.id,
-        tags: this._codeSnippetEditorMetaData.selectedTags
-      })
-    });
-
     // remove the dirty state
     this.title.className = this.title.className.replace(
       ` ${EDITOR_DIRTY_CLASS}`,
@@ -484,7 +452,7 @@ export class CodeSnippetEditor extends ReactWidget {
     }
 
     // update the display in code snippet explorer
-    this.codeSnippetWidget.updateCodeSnippets();
+    this.codeSnippetWidget.updateCodeSnippetWidget();
 
     // close editor if it's from scratch
     if (this._codeSnippetEditorMetaData.fromScratch) {
@@ -499,7 +467,7 @@ export class CodeSnippetEditor extends ReactWidget {
     }
 
     this._codeSnippetEditorMetaData.selectedTags = selectedTags;
-    this._codeSnippetEditorMetaData.allTags = allTags;
+    this._codeSnippetEditorMetaData.allSnippetTags = allTags;
 
     this.saved = false;
   }
@@ -542,7 +510,7 @@ export class CodeSnippetEditor extends ReactWidget {
           required
         />
         <datalist id="languages">
-          {SUPPORTED_LANGUAGES.map(lang => this.renderLanguageOptions(lang))}
+          {SUPPORTED_LANGUAGES.map((lang) => this.renderLanguageOptions(lang))}
         </datalist>
       </div>
     );
@@ -571,10 +539,9 @@ export class CodeSnippetEditor extends ReactWidget {
           <input
             className={CODE_SNIPPET_EDITOR_NAME_INPUT}
             defaultValue={this._codeSnippetEditorMetaData.name}
-            placeholder={'Ex. starter_code'}
+            placeholder={'Ex. starter code'}
             type="text"
             required
-            pattern={'[a-zA-Z0-9_]+'}
             onMouseDown={(
               event: React.MouseEvent<HTMLInputElement, MouseEvent>
             ): void => this.activeFieldState(event)}
@@ -583,21 +550,14 @@ export class CodeSnippetEditor extends ReactWidget {
             }}
             onBlur={this.handleOnBlur}
           ></input>
-          <p className={CODE_SNIPPET_EDITOR_INPUTNAME_VALIDITY}>
-            {
-              'Name of the code snippet MUST be lowercased, alphanumeric or composed of underscore(_)'
-            }
-          </p>
           <label className={CODE_SNIPPET_EDITOR_LABEL}>
-            Description (required)
+            Description (optional)
           </label>
           <input
             className={CODE_SNIPPET_EDITOR_DESC_INPUT}
             defaultValue={this._codeSnippetEditorMetaData.description}
             placeholder={'Description'}
             type="text"
-            required
-            pattern={'[a-zA-Z0-9_ ,.?!]+'}
             onMouseDown={(
               event: React.MouseEvent<HTMLInputElement, MouseEvent>
             ): void => this.activeFieldState(event)}
@@ -606,11 +566,6 @@ export class CodeSnippetEditor extends ReactWidget {
             }}
             onBlur={this.handleOnBlur}
           ></input>
-          <p className={CODE_SNIPPET_EDITOR_INPUTDESC_VALIDITY}>
-            {
-              'Description of the code snippet MUST be alphanumeric but can include space or punctuation'
-            }
-          </p>
           <label className={CODE_SNIPPET_EDITOR_LABEL}>
             Language (required)
           </label>
@@ -618,7 +573,8 @@ export class CodeSnippetEditor extends ReactWidget {
           <label className={CODE_SNIPPET_EDITOR_LABEL}>Tags</label>
           <CodeSnippetEditorTags
             selectedTags={this.codeSnippetEditorMetadata.selectedTags}
-            tags={this.codeSnippetEditorMetadata.allTags}
+            snippetTags={this.codeSnippetEditorMetadata.allSnippetTags}
+            langTags={this.codeSnippetEditorMetadata.allLangTags}
             handleChange={this.handleChangeOnTag}
           />
         </section>
